/* math_3.c 
 * Miscellaneous math functions for RLaB */

/*  This file is a part of RLaB ("Our"-LaB)
   Copyright (C) 1992, 1993, 1994  Ian R. Searle

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   See the file ./COPYING
   ********************************************************************** */

#include "rlab.h"
#include "mem.h"
#include "bltin.h"
#include "symbol.h"
#include "util.h"
#include "matop1.h"
#include "matop2.h"
#include "listnode.h"
#include "btree.h"
#include "r_string.h"
#include "fi_1.h"
#include "fi_2.h"
#include "mathl.h"

#include <math.h>

#define rabs(x) ((x) >= 0 ? (x) : -(x))
#define min(a,b) ((a) <= (b) ? (a) : (b))
#define max(a,b) ((a) >= (b) ? (a) : (b))

/* **************************************************************
 * Return the REAL part of a complex entity.
 * ************************************************************** */
void
Real (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  ListNode *M;

  /* Check n_args */
  if (n_args != 1)
  {
    clean_bltin_args (d_arg, n_args);
    error_1 ("real: 1 argument allowed", 0);
  }

  M = bltin_get_numeric_matrix ("real", d_arg, 1);
  *return_ptr = (VPTR) matrix_Real (e_data (M));

  remove_tmp_destroy (M);
  return;
}

void
Imag (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  ListNode *M;

  /* Check n_args */
  if (n_args != 1)
  {
    clean_bltin_args (d_arg, n_args);
    error_1 ("imag: 1 argument allowed", 0);
  }

  M = bltin_get_numeric_matrix ("imag", d_arg, 1);
  *return_ptr = (VPTR) matrix_Imag (e_data (M));
  remove_tmp_destroy (M);
  return;
}

void
Conj (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  ListNode *M;
  Matrix *m;

  /* Check n_args */
  if (n_args != 1)
  {
    clean_bltin_args (d_arg, n_args);
    error_1 ("conj: 1 argument allowed", 0);
  }

  M = bltin_get_numeric_matrix ("conj", d_arg, 1);
  m = matrix_Conj (e_data (M));
  *return_ptr = (VPTR) m;
  remove_tmp_destroy (M);
  return;
}

#include "fftp.h"

void
FFT (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  F_INT n;
  int i, N, vtype;
  ListNode *X;
  ListNode *tmp1, *tmp2;
  Matrix *m, *mfft, *wsave;

  /* Check n_args */
  if (n_args != 1 && n_args != 2)
  {
    clean_bltin_args (d_arg, n_args);
    error_1 ("fft: 1 or 2 arguments allowed", 0);
  }

  /* get arg from list */
  X = bltin_get_numeric_matrix ("fft", d_arg, 1);

  if (n_args == 2)
  {
    N = bltin_get_numeric_double ("fft", d_arg, 2);
  }
  else
    N = 0;

  m = (Matrix *) e_data (X);
  if (MNR (m) == 1 || MNC (m) == 1)
  {
    /*
     * Vector fft
     */

    if (MNR (m) > MNC (m))	/* Set n, and figure out vector type */
    {
      n = (F_INT) MNR (m);
      vtype = 1;
    }
    else
    {
      n = (F_INT) MNC (m);
      vtype = 2;
    }

    tmp1 = install_tmp (MATRIX, mfft = matrix_copy_complex (m),
			matrix_Destroy);

    if (N != 0)
    {
      /* Set length of vector to N */
      if (N > (int) n)
      {
	if (vtype == 1)		/* column vector */
	  matrix_Extend (mfft, N, 1);
	else			/* row vector */
	  matrix_Extend (mfft, 1, N);
	n = (F_INT) N;
      }
      else if (N < (int) n)
      {
	if (vtype == 1)		/* column vector */
	  matrix_Truncate (mfft, N, 1);
	else			/* row vector */
	  matrix_Truncate (mfft, 1, N);
	n = (F_INT) N;
      }
    }

    tmp2 = install_tmp (MATRIX, wsave = matrix_Create (1, 4 * n + 15),
			matrix_Destroy);

    signal (SIGINT, intcatch);
    CFFTI (&n, MDPTRr (wsave));
    CFFTF (&n, (double *) MDPTRc (mfft), MDPTRr (wsave));
    signal (SIGINT, intcatch_wait);
  }
  else
  {
    /* MAtrix fft, one column at a time */
    n = (F_INT) MNR (m);

    tmp1 = install_tmp (MATRIX, mfft = matrix_copy_complex (m),
			matrix_Destroy);

    if (N != 0)
    {
      /* Set row dimension of [mfft] to N */
      if (N > (int) n)
      {
	matrix_Extend (mfft, N, MNC (mfft));
	n = (F_INT) N;
      }
      else if (N < (int) n)
      {
	matrix_Truncate (mfft, N, MNC (mfft));
	n = (F_INT) N;
      }
    }

    tmp2 = install_tmp (MATRIX, wsave = matrix_Create (1, 4 * n + 15),
			matrix_Destroy);

    for (i = 0; i < MNC (m); i++)
    {
      signal (SIGINT, intcatch);
      CFFTI (&n, MDPTRr (wsave));
      CFFTF (&n, (double *) (mfft->val.mc + (i * n)), MDPTRr (wsave));
      signal (SIGINT, intcatch_wait);
    }
  }
  remove_tmp (tmp1);
  remove_tmp_destroy (tmp2);
  remove_tmp_destroy (X);
  *return_ptr = (VPTR) mfft;
}

void
IFFT (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  F_INT n;
  int i, N, vtype;
  ListNode *X;
  ListNode *tmp1, *tmp2;
  Matrix *m, *mfft, *wsave;

  /* Check n_args */
  if (n_args != 1 && n_args != 2)
  {
    clean_bltin_args (d_arg, n_args);
    error_1 ("ifft: 1 or 2 arguments allowed", 0);
  }

  X = bltin_get_numeric_matrix ("ifft", d_arg, 1);

  if (n_args == 2)
  {
    N = (int) bltin_get_numeric_double ("ifft", d_arg, 2);
  }
  else
    N = 0;
  
  m = (Matrix *) e_data (X);
  if (MNR (m) == 1 || MNC (m) == 1)
  {
    /*
     * Vector fft
     */

    if (MNR (m) > MNC (m))	/* Set n, and figure out vector type */
    {
      n = (F_INT) MNR (m);
      vtype = 1;
    }
    else
    {
      n = (F_INT) MNC (m);
      vtype = 2;
    }

    tmp1 = install_tmp (MATRIX, mfft = matrix_copy_complex (m),
			matrix_Destroy);

    if (N != 0)
    {
      /* Set length of vector to N */
      if (N > (int) n)
      {
	if (vtype == 1)		/* column vector */
	  matrix_Extend (mfft, N, 1);
	else			/* row vector */
	  matrix_Extend (mfft, 1, N);
	n = (F_INT) N;
      }
      else if (N < (int) n)
      {
	if (vtype == 1)		/* column vector */
	  matrix_Truncate (mfft, N, 1);
	else			/* row vector */
	  matrix_Truncate (mfft, 1, N);
	n = (F_INT) N;
      }
    }

    tmp2 = install_tmp (MATRIX, wsave = matrix_Create (1, 4 * n + 15),
			matrix_Destroy);

    signal (SIGINT, intcatch);
    CFFTI (&n, MDPTRr (wsave));
    CFFTB (&n, (double *) MDPTRc (mfft), MDPTRr (wsave));
    signal (SIGINT, intcatch_wait);
  }
  else
  {
    /* Matrix ifft, one column at a time */
    n = MNR (m);

    tmp1 = install_tmp (MATRIX, mfft = matrix_copy_complex (m),
			matrix_Destroy);

    if (N != 0)
    {
      /* Set row dimension of [mfft] to N */
      if (N > (int) n)
      {
	matrix_Extend (mfft, N, MNC (mfft));
	n = (F_INT) N;
      }
      else if (N < (int) n)
      {
	matrix_Truncate (mfft, N, MNC (mfft));
	n = (F_INT) N;
      }
    }

    tmp2 = install_tmp (MATRIX, wsave = matrix_Create (1, 4 * n + 15),
			matrix_Destroy);

    for (i = 0; i < MNC (m); i++)
    {
      signal (SIGINT, intcatch);
      CFFTI (&n, MDPTRr (wsave));
      CFFTB (&n, (double *) (mfft->val.mc + (i * n)), MDPTRr (wsave));
      signal (SIGINT, intcatch_wait);
    }
  }

  /*
   * Scale the result by 1/N
   */
  for (i = 0; i < MNR (mfft) * MNC (mfft); i++)
  {
    MATcvr (mfft, i) = MATcvr (mfft, i) / ((double) n);
    MATcvi (mfft, i) = MATcvi (mfft, i) / ((double) n);
  }

  remove_tmp (tmp1);
  remove_tmp_destroy (tmp2);
  remove_tmp_destroy (X);
  *return_ptr = (VPTR) mfft;
}

/*
 * Sum the elements of a matrix
 */

void
Sum (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  ListNode *M;

  /* Check n_args */
  if (n_args != 1)
  {
    clean_bltin_args (d_arg, n_args);
    error_1 ("sum: 1 argument allowed", 0);
  }

  M = bltin_get_numeric_matrix ("sum", d_arg, 1);
  *return_ptr = (VPTR) matrix_Sum (e_data (M));
  remove_tmp_destroy (M);
  return;
}

void
CumSum (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  ListNode *M;

  /* Check n_args */
  if (n_args != 1)
  {
    clean_bltin_args (d_arg, n_args);
    error_1 ("cumsum: 1 argument allowed", 0);
  }

  M = bltin_get_numeric_matrix ("cumsum", d_arg, 1);
  *return_ptr = (VPTR) matrix_CumSum (e_data (M));
  remove_tmp_destroy (M);
  return;
}

void
Prod (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  ListNode *M;

  /* Check n_args */
  if (n_args != 1)
    error_1 ("prod: 1 argument allowed", 0);

  M = bltin_get_numeric_matrix ("prod", d_arg, 1);
  *return_ptr = (VPTR) matrix_Prod (e_data (M));
  remove_tmp_destroy (M);
  return;
}

void
CumProd (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  ListNode *M;

  /* Check n_args */
  if (n_args != 1)
    error_1 ("cumprod: 1 argument allowed", 0);

  M = bltin_get_numeric_matrix ("cumprod", d_arg, 1);
  *return_ptr = (VPTR) matrix_CumProd (e_data (M));
  remove_tmp_destroy (M);
  return;
}

/* **************************************************************
 * Factor, and Backsub functions.
 * ************************************************************** */

extern int matrix_is_symm _PROTO ((Matrix *m));

void
Factor (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  char *str;
  double rcond;
  int flag;
  Btree *rlist = 0;
  ListNode *A, *TYPE;
  Matrix *lu, *pvt;
  Matrix *stmp;

  /* Check n_args */
  if (!(n_args == 1 || n_args == 2))
    error_1 ("factor: 1 or 2 argument(s) allowed", 0);

  A = bltin_get_numeric_matrix ("factor", d_arg, 1);

  flag = 0;
  if (n_args == 2)
  {
    TYPE = bltin_get_string ("factor", d_arg, 2);
    str = string_GetString (e_data (TYPE));
    if (!strncmp ("s", str, 1))
      flag = 2;
    else if (!strncmp ("S", str, 1))
      flag = 2;
    else if (!strncmp ("g", str, 1))
      flag = 1;
    else if (!strncmp ("G", str, 1))
      flag = 1;
    else
      error_1 ("factor: Invalid 2nd argument", 0);
  }

  /*
   * Check symmetry
   */

  if (flag ==0)
  {
    if (matrix_is_symm (e_data (A)))
    {
      matrix_Factor_Sy (e_data (A), &lu, &pvt, &rcond);
      rlist = btree_Create ();
      install (rlist, cpstr ("ldl"), MATRIX, lu);
      install (rlist, cpstr ("pvt"), MATRIX, pvt);
      install (rlist, cpstr ("rcond"), MATRIX, 
	       stmp = matrix_CreateScalarR (rcond));
      matrix_SetName (lu, cpstr ("ldl"));
      matrix_SetName (pvt, cpstr ("pvt"));
      matrix_SetName (stmp, cpstr ("rcond"));
    }
    else
    {
      matrix_Factor_Ge (e_data (A), &lu, &pvt, &rcond);
      rlist = btree_Create ();
      install (rlist, cpstr ("lu"), MATRIX, lu);
      install (rlist, cpstr ("pvt"), MATRIX, pvt);
      install (rlist, cpstr ("rcond"), MATRIX, 
	       stmp = matrix_CreateScalarR (rcond));
      matrix_SetName (lu, cpstr ("lu"));
      matrix_SetName (pvt, cpstr ("pvt"));
      matrix_SetName (stmp, cpstr ("rcond"));
    }
  }
  else if (flag == 1)
  {
    /* User is forcing general solution */
    matrix_Factor_Ge (e_data (A), &lu, &pvt, &rcond);
    rlist = btree_Create ();
    install (rlist, cpstr ("lu"), MATRIX, lu);
    install (rlist, cpstr ("pvt"), MATRIX, pvt);
    install (rlist, cpstr ("rcond"), MATRIX, 
	     stmp = matrix_CreateScalarR (rcond));
    matrix_SetName (lu, cpstr ("lu"));
    matrix_SetName (pvt, cpstr ("pvt"));
    matrix_SetName (stmp, cpstr ("rcond"));
  }
  else if (flag == 2)
  {
    /* User if forcing symmetric solution */
    matrix_Factor_Sy (e_data (A), &lu, &pvt, &rcond);
    rlist = btree_Create ();
    install (rlist, cpstr ("ldl"), MATRIX, lu);
    install (rlist, cpstr ("pvt"), MATRIX, pvt);
    install (rlist, cpstr ("rcond"), MATRIX, 
	     stmp = matrix_CreateScalarR (rcond));
    matrix_SetName (lu, cpstr ("ldl"));
    matrix_SetName (pvt, cpstr ("pvt"));
    matrix_SetName (stmp, cpstr ("rcond"));
  }
  
  *return_ptr = (VPTR) rlist;
  remove_tmp_destroy (A);
  return;
}

void
Backsub (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  ListNode *F, *B;
  Matrix *sol;

  /* Check n_args */
  if (n_args != 2)
    error_1 ("backsub: requires 2 arguments", 0);

  F = bltin_get_list ("backsub", d_arg, 1);
  B = bltin_get_numeric_matrix ("backsub", d_arg, 2);
  sol = matrix_Backsub (e_data (F), e_data (B));
  *return_ptr = (VPTR) sol;
  remove_tmp_destroy (F);
  remove_tmp_destroy (B);
  return;
}

void
Schur (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  Btree *rlist;
  ListNode *M;
  Matrix *t, *z;

  /* Check n_args */
  if (n_args != 1)
    error_1 ("schur: 1 argument allowed", 0);

  M = bltin_get_numeric_matrix ("schur", d_arg, 1);
  matrix_Schur (e_data (M), &t, &z);
  
  rlist = btree_Create ();
  install (rlist, cpstr ("t"), MATRIX, t);
  install (rlist, cpstr ("z"), MATRIX, z);
  matrix_SetName (t, cpstr ("t"));
  matrix_SetName (z, cpstr ("z"));

  *return_ptr = (VPTR) rlist;
  remove_tmp_destroy (M);
  return;
}

void
Sylv (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  ListNode *A, *B, *C;
  Matrix *a, *b, *c, *x;

  /* Check n_args */
  if (!(n_args == 2 || n_args == 3))
    error_1 ("sylv: 2 or 3 arguments are required", 0);

  /* get arg from list */
  if (n_args == 2)
  {
    /* Lyapunov */
    A = bltin_get_numeric_matrix ("sylv", d_arg, 1);
    C = bltin_get_numeric_matrix ("sylv", d_arg, 2);

    a = (Matrix *) e_data (A);
    c = (Matrix *) e_data (C);

    matrix_Sylv (a, (Matrix *) 0, c, &x);

    *return_ptr = (VPTR) x;

    remove_tmp_destroy (A);
    remove_tmp_destroy (C);
    return;
  }
  else if (n_args == 3)
  {
    /* Sylvester */
    A = bltin_get_numeric_matrix ("sylv", d_arg, 1);
    B = bltin_get_numeric_matrix ("sylv", d_arg, 2);
    C = bltin_get_numeric_matrix ("sylv", d_arg, 3);

    a = (Matrix *) e_data (A);
    b = (Matrix *) e_data (B);
    c = (Matrix *) e_data (C);

    matrix_Sylv (a, b, c, &x);

    *return_ptr = (VPTR) x;

    remove_tmp_destroy (A);
    remove_tmp_destroy (B);
    remove_tmp_destroy (C);
    return;
  }
}

/* **************************************************************
 * Functions for determining symmetry
 * ************************************************************** */

int matrix_is_symm _PROTO ((Matrix * m));

void
Issymm (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  int retval;
  ListNode *M;

  /* Check n_args */
  if (n_args != 1)
    error_1 ("issym: 1 argument allowed", 0);

  M = bltin_get_matrix ("issym", d_arg, 1);
  retval = matrix_is_symm (e_data (M));
  
  *return_ptr = (VPTR) matrix_CreateScalarR ((double) retval);
  
  remove_tmp_destroy (M);
  return;
}

/*
 * Return TRUE (1) if the input matrix is symmetric (Hermitian),
 * FALSE (0) otherwise. Check the matrix colum-wise.
 * Return FALSE (0) at the 1st failed test.
 */

int
matrix_is_symm (m)
     Matrix *m;
{
  int i, j, ndiag;

  if (MNR (m) != MNC (m))
    return (0);

  if (MTYPE (m) == REAL)
  {
    for (j = 0; j < MNC (m); j++)
    {
      for (i = j + 1; i < MNR (m); i++)
      {
	if (MAT0 (m, i, j) != MAT0 (m, j, i))
	  return (0);
      }
    }
  }
  else if (MTYPE (m) == COMPLEX)
  {
    /*
     * 1st check the diagonal elements to make 
     * sure they are all REAL.
     */

    ndiag = min (MNR (m), MNC (m));
    for (i = 0; i < ndiag; i++)
    {
      if (MATi0 (m, i, i) != 0.0)
	return (0);
    }

    /*
     * Now check the off-diagonal elements.
     */

    for (j = 0; j < MNC (m); j++)
    {
      for (i = j + 1; i < MNR (m); i++)
      {
	if (MATr0 (m, i, j) != MATr0 (m, j, i))
	  return (0);
	if (MATi0 (m, i, j) != -MATi0 (m, j, i))
	  return (0);
      }
    }
  }
  else if (MTYPE (m) == STRING)
  {
    for (j = 0; j < MNC (m); j++)
    {
      for (i = j + 1; i < MNR (m); i++)
      {
	if (strcmp (MATs0 (m, i, j), MATs0 (m, j, i)))
	  return (0);
      }
    }
  }

  return (1);
}

/* **************************************************************
 * Symmetric Eigenvalue solvers
 * ************************************************************** */
void
EigS (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  int sol = 0;
  Btree *rlist;
  ListNode *A, *B;
  Matrix *vec, *val;

  /* Check n_args */
  if ((n_args != 1) && (n_args != 2))
    error_1 ("eigs: 1 or 2 arguments allowed", 0);

  A = B = 0;  /* Initialize */

  if (n_args == 1)
  {
    A = bltin_get_numeric_matrix ("eigs", d_arg, 1);
    sol = 1;			/* Standard symmetric problem */
  }
  else if (n_args == 2)
  {
    A = bltin_get_numeric_matrix ("eigs", d_arg, 1);
    B = bltin_get_numeric_matrix ("eigs", d_arg, 2);
    sol = 2;			/* Generalized symmetric problem */
  }
  else
    error_1 ("eig: invalid number of arguments", 0);

  /* Create list for returning results */
  rlist = btree_Create ();

  if (sol == 1)
  {
    /* Standard Eigenvalue Problem, Symmetric */
    matrix_Eig_SEP (e_data (A), &val, &vec);
    install (rlist, cpstr ("val"), MATRIX, val);
    install (rlist, cpstr ("vec"), MATRIX, vec);
    matrix_SetName (val, cpstr ("val"));
    matrix_SetName (vec, cpstr ("vec"));
    remove_tmp_destroy (A);
  }
  else if (sol == 2)
  {
    /* Generalized problem, symmetric */
    matrix_Eig_GSEP (e_data (A), e_data (B),
		     &val, &vec);
    install (rlist, cpstr ("val"), MATRIX, val);
    install (rlist, cpstr ("vec"), MATRIX, vec);
    matrix_SetName (val, cpstr ("val"));
    matrix_SetName (vec, cpstr ("vec"));
    remove_tmp_destroy (A);
    remove_tmp_destroy (B);
  }
  *return_ptr = (VPTR) rlist;
}

/* **************************************************************
 * Non-Symmetric Eigenvalue solvers
 * ************************************************************** */
void
EigN (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  int sol = 0;
  Btree *rlist;
  ListNode *A, *B;
  Matrix *val, *lvec, *rvec, *alpha, *beta;

  /* Check n_args */
  if ((n_args != 1) && (n_args != 2))
    error_1 ("eign: 1 or 2 arguments allowed", 0);

  A = B = 0;  /* Initialize */

  if (n_args == 1)
  {
    A = bltin_get_numeric_matrix ("eigs", d_arg, 1);
    sol = 1;			/* Standard non-symmetric problem */
  }
  else if (n_args == 2)
  {
    A = bltin_get_numeric_matrix ("eigs", d_arg, 1);
    B = bltin_get_numeric_matrix ("eigs", d_arg, 2);
    sol = 2;			/* Generalized non-symmetric problem */
  }
  else
    error_1 ("eign: invalid number of arguments", 0);

  /* Create list for returning results */
  rlist = btree_Create ();

  if (sol == 1)
  {
    /* Standard Eigenvalue Problem, Non-Symmetric */
    matrix_Eig_NEP (e_data (A), &val, &rvec, &lvec, 1);

    remove_tmp_destroy (A);
    install (rlist, cpstr ("val"), MATRIX, val);
    install (rlist, cpstr ("rvec"), MATRIX, rvec);
    install (rlist, cpstr ("lvec"), MATRIX, lvec);
    matrix_SetName (val, cpstr ("val"));
    matrix_SetName (rvec, cpstr ("rvec"));
    matrix_SetName (lvec, cpstr ("lvec"));
  }
  else if (sol == 2)
  {
    /* Generalized problem, Non-symmetric */
    matrix_Eig_GNEPa (e_data (A), e_data (B),
		      &alpha, &beta, &rvec, &lvec);

    install (rlist, cpstr ("alpha"), MATRIX, alpha);
    install (rlist, cpstr ("beta"), MATRIX, beta);
    install (rlist, cpstr ("rvec"), MATRIX, rvec);
    install (rlist, cpstr ("lvec"), MATRIX, lvec);
    matrix_SetName (alpha, cpstr ("alpha"));
    matrix_SetName (beta, cpstr ("beta"));
    matrix_SetName (rvec, cpstr ("rvec"));
    matrix_SetName (lvec, cpstr ("lvec"));
    remove_tmp_destroy (A);
    remove_tmp_destroy (B);
  }

  *return_ptr = (VPTR) rlist;
}

void
SchOrd (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  Btree *rlist;
  ListNode *L, *tmp;
  Matrix *t, *z, *T, *Z;
  int ifst, ilst;

  /* Check n_args */
  if (n_args != 3)
    error_1 ("schord: 3 arguments required", 0);

  L = bltin_get_list ("schord", d_arg, 1);
  ifst = (int) bltin_get_numeric_double ("schord", d_arg, 2);
  ilst = (int) bltin_get_numeric_double ("schord", d_arg, 3);

  if (!(tmp = btree_FindNode (e_data (L), "t")))
    error_1 ("List must contain \"t\" element for schord()", 0);
  t = (Matrix *) e_data (tmp);

  if (!(tmp = btree_FindNode (e_data (L), "z")))
    error_1 ("List must contain \"z\" element for schord()", 0);
  z = (Matrix *) e_data (tmp);

  /* Check correctness of ifst, and ilst arguments */
  if (ifst < 1 || ilst < 1)
    error_1 ("schord: IFST, and ILST must be > 0", 0);
  if (ifst > MNR (z) || ilst > MNR (z))
    error_1 ("schord: IFST and ILST must be <= z.n", 0);

  matrix_Schur_Reorder (t, z, ifst, ilst, &T, &Z);

  rlist = btree_Create ();
  install (rlist, cpstr ("t"), MATRIX, T);
  install (rlist, cpstr ("z"), MATRIX, Z);
  matrix_SetName (t, cpstr ("t"));
  matrix_SetName (z, cpstr ("z"));

  *return_ptr = (VPTR) rlist;
  remove_tmp_destroy (L);
  return;
}

/* **************************************************************
 * Filter: 
 * ************************************************************** */

static void filter _PROTO ((double *b, double *a, double *x,
			    double *y, double *v, double *vi,
			    int *ntotal, int *M));

void
Filter (return_ptr, n_args, d_arg)
     VPTR *return_ptr;
     int n_args;
     Datum *d_arg;
{
  int M, N, NN, i, ntotal;
  double *v, *vi;
  Btree *rlist;
  ListNode *tmp1, *tmp2;
  ListNode *A, *B, *X, *ZI;
  Matrix *a, *b, *x, *y, *zi, *zf;
  Matrix *aa, *bb;

  /* Check n_args */
  if (n_args < 3)
    error_1 ("filter: requires at least 3 arguments", 0);

  ZI = 0;  /* Initialize */

  /*
   * Get args from list and error-check.
   */

  B = bltin_get_numeric_matrix ("filter", d_arg, 1);
  A = bltin_get_numeric_matrix ("filter", d_arg, 2);
  X = bltin_get_numeric_matrix ("filter", d_arg, 3);

  a = (Matrix *) e_data (A);
  b = (Matrix *) e_data (B);
  x = (Matrix *) e_data (X);

  if (MTYPE (a) != REAL)
    error_1 ("filter: A must be type: REAL", 0);
  if (MTYPE (b) != REAL)
    error_1 ("filter: B must be type: REAL", 0);
  if (MTYPE (x) != REAL)
    error_1 ("filter: X must be type: REAL", 0);

  if (MNR (x) != 1 && MNC (x) != 1)
    error_1 ("filter: X must be a row or column matrix", 0);
  if (MNR (a) != 1 && MNC (a) != 1)
    error_1 ("filter: A must be a row or column matrix", 0);
  if (MNR (b) != 1 && MNC (b) != 1)
    error_1 ("filter: B must be a row or column matrix", 0);

  ntotal = MNR (x) * MNC (x);
  M = MNR (b) * MNC (b);
  N = MNR (a) * MNC (a);
  NN = max (M, N);
  y = matrix_Create (MNR (x), MNC (x));

  /*
   * Fix up pole and zero vectors.
   * Make them the same length, this makes
   * filter's job much easier.
   */

  tmp1 = install_tmp (MATRIX, aa = matrix_Create (NN, 1),
		      matrix_Destroy);
  tmp2 = install_tmp (MATRIX, bb = matrix_Create (NN, 1),
		      matrix_Destroy);
  matrix_Zero (aa);
  matrix_Zero (bb);

  if (MATrv (a, 0) == 0)
    error_1 ("filter: 1st A term must be non-zero", 0);

  for (i = 0; i < M; i++)
    MATrv (bb, i) = MATrv (b, i) / MATrv (a, 0);
  for (i = 1; i < N; i++)
    MATrv (aa, i) = MATrv (a, i) / MATrv (a, 0);

  /*
   * Create delay vectors and load inital delays.
   * Add an extra term to vi[] to make filter's 
   * job a little easier. This extra term will
   * always be zero.
   */

  v = (double *) MALLOC (sizeof (double) * NN);
  vi = (double *) MALLOC (sizeof (double) * (NN + 1));

  if (n_args == 4)
  {
    ZI = bltin_get_numeric_matrix ("filter", d_arg, 4);
    zi = (Matrix *) e_data (ZI);
    if (MTYPE (zi) != REAL)
      error_1 ("filter: Zi must be type: REAL", 0);
    if (MNR (zi) * MNC (zi) != NN)
      error_1 ("filter: Zi must be max(size(A), size(B))", 0);
    for (i = 0; i < NN; i++)
      vi[i] = MATrv (zi, i);
    vi[NN] = 0.0;
  }
  else
  {
    for (i = 0; i < NN; i++)
      vi[i] = 0.0;
    vi[NN] = 0.0;
  }

  /*
   * Do the work...
   */

  filter (MDPTRr (bb), MDPTRr (aa), MDPTRr (x), MDPTRr (y),
	  v, vi, &ntotal, &NN);

  zf = matrix_Create (NN, 1);
  for (i = 0; i < NN; i++)
    MATrv (zf, i) = v[i];

  rlist = btree_Create ();
  install (rlist, cpstr ("y"), MATRIX, y);
  install (rlist, cpstr ("zf"), MATRIX, zf);
  matrix_SetName (y, cpstr ("y"));
  matrix_SetName (zf, cpstr ("zf"));

  *return_ptr = (VPTR) rlist;

  FREE (v);
  FREE (vi);

  remove_tmp_destroy (tmp1);
  remove_tmp_destroy (tmp2);

  remove_tmp_destroy (A);
  remove_tmp_destroy (B);
  remove_tmp_destroy (X);
  if (n_args == 4)
    remove_tmp_destroy (ZI);

  return;
}

/*
 * This is the function that acutally computes the
 * response. There is no error checking done here,
 * because Filter() does it all.
 */

static void
filter (b, a, x, y, v, vi, ntotal, N)
     double *b, *a, *x, *y, *v, *vi;
     int *ntotal, *N;
{
  int k, n;

  for (n = 0; n < *ntotal; n++)
  {
    v[0] = b[0] * x[n] + vi[1];
    y[n] = v[0];
    for (k = 1; k < *N; k++)
    {
      v[k] = b[k] * x[n] - a[k] * v[0] + vi[k + 1];
      vi[k] = v[k];
    }
  }
}
